@@ -47,7 +47,7 @@
 #include <pcap.h>
 #include "list.h"
 
-static char version[] = "0.3.1";
+static char version[] = "0.4";
 
 #define SEQ_TIMEOUT 25 /* default knock timeout in seconds */
 #define CMD_TIMEOUT 10 /* default timeout in seconds between start and stop commands */
@@ -78,7 +78,8 @@ PMList *doors = NULL;
 typedef struct knocker {
 	opendoor_t *door;
 	short stage;
-	char src[16];
+	char src[16];   /* IP address */
+	char *srchost;  /* Hostname */
 	time_t seq_start;
 } knocker_t;
 PMList *attempts = NULL;
@@ -97,6 +98,7 @@ char* trim(char *str);
 void runCommand(char *cmd);
 int parseconfig(char *configfile);
 int get_mac(const char* iface, u_char *addr);
+char* get_ip(const char* iface, char *buf, int bufsize);
 char* parse_cmd(char* command, char* src);
 int exec_cmd(char* command, char* name);
 void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet);
@@ -104,6 +106,7 @@ void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet);
 pcap_t *cap = NULL;
 FILE *logfd = NULL;
 int lltype = -1;
+char myip[32];
 u_char if_mac[6];
 
 int  o_usesyslog = 0;
@@ -178,15 +181,35 @@ int main(int argc, char **argv)
 		case DLT_LINUX_SLL:
 			dprint("ppp interface detected (linux \"cooked\" encapsulation)\n");
 			break;
+		case DLT_RAW:
+			dprint("raw interface detected, no encapsulation\n");
+			break;
 		default: 
 			fprintf(stderr, "error: unsupported link-layer type: %d\n", lltype);
 			cleanup(1);
 			break;
 	}
 
-	if(lltype == DLT_EN10MB && get_mac(o_int, if_mac)) {
-		fprintf(stderr, "could not get MAC address for %s\n", o_int);
+	if(lltype == DLT_EN10MB) {
+		if(get_mac(o_int, if_mac)) {
+			fprintf(stderr, "could not get MAC address for %s\n", o_int);
+			cleanup(1);
+		} else {
+			int i;
+			dprint("MAC address: ");
+			for(i = 0; i < 6; i++) {
+				dprint("%s%02X", i>0 ? ":" : "", if_mac[i]);
+			}
+			dprint("\n");
+		}
+	}
+
+	/* get our local IP address */
+	if(get_ip(o_int, myip, 32) == NULL) {
+		fprintf(stderr, "could not get IP address for %s\n", o_int);
 		cleanup(1);
+	} else {
+		dprint("Local IP: %s\n", myip);
 	}
 
 	if(o_daemon) {
@@ -266,8 +289,8 @@ void logprint(char *fmt, ...)
 		t = time(NULL);
 		tm = localtime(&t);
 		
-		fprintf(logfd, "[%02d/%02d/%02d %02d:%02d] %s\n", tm->tm_mon+1, tm->tm_mday,
-			tm->tm_year-100, tm->tm_hour, tm->tm_min, msg);
+		fprintf(logfd, "[%04d-%02d-%02d %02d:%02d] %s\n", tm->tm_year+1900,
+			tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min, msg);
 		fflush(logfd);
 	}
 }
@@ -320,7 +343,7 @@ void usage() {
 
 void ver() {
 	printf("knockd %s\n", version);
-	printf("Copyright (C) 2004 Judd Vinet <jvinet@zeroflux.org>\n");
+	printf("Copyright (C) 2004-2005 Judd Vinet <jvinet@zeroflux.org>\n");
 	exit(0);
 }
 
@@ -595,6 +618,40 @@ int get_mac(const char* iface, u_char *addr)
 	return(1);
 }
 
+/* Get the IP address of an interface
+ */
+char* get_ip(const char* iface, char *buf, int bufsize)
+{
+	int s;
+	struct ifreq ifr;
+
+	if(bufsize <= 0) {
+		return(NULL);
+	}
+	if(buf == NULL) {
+		return(NULL);
+	}
+	buf[0] = '\0';
+
+	s = socket(AF_INET, SOCK_DGRAM, 0);
+	if(s < 0) {
+		return(NULL);
+	}
+
+	bzero((void*)(&ifr.ifr_name), sizeof(ifr.ifr_name));
+	strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name));
+	if(ioctl(s, SIOCGIFADDR, &ifr)) {
+		close(s);
+		return(NULL);
+	}
+	close(s);
+
+	strncpy(buf, inet_ntoa((*(struct sockaddr_in *)&ifr.ifr_addr).sin_addr), bufsize);
+	return(buf);
+}
+
+/* Parse a command line, replacing tokens (eg, %IP%) with their real values
+ */
 char* parse_cmd(char* command, char* src){
 	char *ptr1, *ptr2;
 	char origCmd[PATH_MAX];
@@ -613,6 +670,8 @@ char* parse_cmd(char* command, char* src){
 	return command;
 }
 
+/* Execute a command through the system shell and wait for return
+ */
 int exec_cmd(char* command, char* name){
 	int ret;
 
@@ -657,7 +716,7 @@ void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet)
 		if(ntohs(eth->h_proto) != ETH_P_IP) {
 			return;
 		}
-		/* make sure this packet was sent TO us, not sent BY us */
+		/* make sure this packet was sent TO us, not sent FROM us */
 		for(i = 0; i < 6; i++) {
 			if(eth->h_dest[i] != if_mac[i]) {
 				return;
@@ -666,11 +725,13 @@ void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet)
 		ip = (struct iphdr*)(packet + sizeof(struct ethhdr));
 	} else if(lltype == DLT_LINUX_SLL) {
 		if(packet[0] == 0 && packet[1] == 0) {
-			/* packet was sent TO us, not sent BY us */
+			/* packet was sent TO us, not sent FROM us */
 			ip = (struct iphdr*)((u_char*)packet + 16);
 		} else {
 			return;
 		}
+	} else if(lltype == DLT_RAW) {
+		ip = (struct iphdr*)((u_char*)packet);
 	}
 	
 	if(ip->version != 4) {
@@ -679,8 +740,20 @@ void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet)
 		return;
 	}
 	if(ip->protocol == IPPROTO_ICMP) {
+		/* we don't do ICMP */
 		return;
 	}
+
+	/* make sure this packet was sent TO us, not FROM us or THROUGH us */
+	if(inet_aton(myip, &inaddr) == 0) {
+		fprintf(stderr, "error: could not understand IP address: %s\n", myip);
+		return;
+	}
+	if(ip->daddr != inaddr.s_addr) {
+		dprint("packet destined for another host, ignoring...\n");
+		return;
+	}
+	
 	sport = dport = 0;
 	if(ip->protocol == IPPROTO_TCP) {
 		strncpy(proto, "tcp", sizeof(proto));
@@ -724,16 +797,27 @@ void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet)
 			nix = 1;
 		}
 		if(!nix && (pkt_secs - attempt->seq_start) >= attempt->door->seq_timeout) {
-			vprint("%s: %s: sequence timeout (stage %d)\n", attempt->src, attempt->door->name, attempt->stage);
-			logprint("%s: %s: sequence timeout (stage %d)", attempt->src, attempt->door->name, attempt->stage);
+			if(attempt->srchost) {
+				vprint("%s (%s): %s: sequence timeout (stage %d)\n", attempt->src, attempt->srchost,
+						attempt->door->name, attempt->stage);
+				logprint("%s (%s): %s: sequence timeout (stage %d)\n", attempt->src, attempt->srchost,
+						attempt->door->name, attempt->stage);
+			} else {
+				vprint("%s: %s: sequence timeout (stage %d)\n", attempt->src,
+						attempt->door->name, attempt->stage);
+				logprint("%s: %s: sequence timeout (stage %d)\n", attempt->src,
+						attempt->door->name, attempt->stage);
+			}
 			nix = 1;
 		}
 		if(nix) {
+			knocker_t *k = (knocker_t*)lp->data;
 			/* splice this entry out of the list */
 			if(lp->prev) lp->prev->next = lp->next;
 			if(lp->next) lp->next->prev = lp->prev;
 			if(lp == attempts) attempts = NULL;
 			lp->prev = lp->next = NULL;
+			free(k->srchost);
 			list_free(lp);
 			continue;
 		}
@@ -783,11 +867,21 @@ void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet)
 				dport == attempt->door->sequence[attempt->stage]) {
 			/* level up! */
 			attempt->stage++;
-			vprint("%s: %s: Stage %d\n", attempt->src, attempt->door->name, attempt->stage);
-			logprint("%s: %s: Stage %d", attempt->src, attempt->door->name, attempt->stage);
+			if(attempt->srchost) {
+				vprint("%s (%s): %s: Stage %d\n", attempt->src, attempt->srchost, attempt->door->name, attempt->stage);
+				logprint("%s (%s): %s: Stage %d", attempt->src, attempt->srchost, attempt->door->name, attempt->stage);
+			} else {
+				vprint("%s: %s: Stage %d\n", attempt->src, attempt->door->name, attempt->stage);
+				logprint("%s: %s: Stage %d", attempt->src, attempt->door->name, attempt->stage);
+			}
 			if(attempt->stage >= attempt->door->seqcount) {
-				vprint("%s: %s: OPEN SESAME\n", attempt->src, attempt->door->name);
-				logprint("%s: %s: OPEN SESAME", attempt->src, attempt->door->name);
+				if(attempt->srchost) {
+					vprint("%s (%s): %s: OPEN SESAME\n", attempt->src, attempt->srchost, attempt->door->name);
+					logprint("%s (%s): %s: OPEN SESAME", attempt->src, attempt->srchost, attempt->door->name);
+				} else {
+					vprint("%s: %s: OPEN SESAME\n", attempt->src, attempt->door->name);
+					logprint("%s: %s: OPEN SESAME", attempt->src, attempt->door->name);
+				}
 				if(attempt->door->start_command && strlen(attempt->door->start_command)) {
 					/* run the associated command */
 					if(fork() == 0) {
@@ -798,13 +892,19 @@ void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet)
 						/* if stop_command is set, sleep for cmd_timeout and run it*/
 						if(attempt->door->stop_command){
 							sleep(attempt->door->cmd_timeout);
-							vprint("%s: %s: command timeout\n", attempt->src, attempt->door->name);
-							logprint("%s: %s: command timeout", attempt->src, attempt->door->name);
+							if(attempt->srchost) {
+								vprint("%s (%s): %s: command timeout\n", attempt->src, attempt->srchost, attempt->door->name);
+								logprint("%s (%s): %s: command timeout", attempt->src, attempt->srchost, attempt->door->name);
+							} else {
+								vprint("%s: %s: command timeout\n", attempt->src, attempt->door->name);
+								logprint("%s: %s: command timeout", attempt->src, attempt->door->name);
+							}
 							exec_cmd(parse_cmd(attempt->door->stop_command, attempt->src), attempt->door->name);
 						}
 						
 						exit(0); /* exit child */
-					}				}
+					}
+				}
 			}
 		} else if(flagsmatch == 0) {
 			/* TCP flags didn't match -- just ignore this packet, don't
@@ -831,18 +931,32 @@ void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet)
 			}
 
 			if(ip->protocol == door->protocol[0] && dport == door->sequence[0]) {
+				struct hostent *he;
 				/* create a new entry */
 				attempt = (knocker_t*)malloc(sizeof(knocker_t));
+				attempt->srchost = NULL;
 				if(attempt == NULL) {
 					perror("malloc");
 					exit(1);
 				}
 				strcpy(attempt->src, srcIP);
+				/* try a reverse lookup */
+				inaddr.s_addr = ip->saddr;
+				he = gethostbyaddr((void *)&inaddr, sizeof(inaddr), AF_INET);
+				if(he) {
+					attempt->srchost = strdup(he->h_name);
+				}
+
 				attempt->stage = 1;
 				attempt->seq_start = pkt_secs;
 				attempt->door = door;
-				vprint("%s: %s: Stage 1\n", attempt->src, door->name);
-				logprint("%s: %s: Stage 1", attempt->src, door->name);
+				if(attempt->srchost) {
+					vprint("%s (%s): %s: Stage 1\n", attempt->src, attempt->srchost, door->name);
+					logprint("%s (%s): %s: Stage 1", attempt->src, attempt->srchost, door->name);
+				} else {
+					vprint("%s: %s: Stage 1\n", attempt->src, door->name);
+					logprint("%s: %s: Stage 1", attempt->src, door->name);
+				}
 				attempts = list_add(attempts, attempt);
 			}
 		}
